@@ -27,11 +27,13 @@
  *
  */
 
-#include "keadataset.h"
 #include "keaband.h"
+#include "keadataset.h"
 #include "keacopy.h"
+#include "hdf5vfl.h"
 
 #include "libkea/KEACommon.h"
+#include "cpl_vsi_virtual.h"
 
 // Function for converting a libkea type into a GDAL type
 GDALDataType KEA_to_GDAL_Type( kealib::KEADataType ekeaType )
@@ -49,12 +51,22 @@ GDALDataType KEA_to_GDAL_Type( kealib::KEADataType ekeaType )
         case kealib::kea_32int:
             egdalType = GDT_Int32;
             break;
+#ifdef HAVE_64BITIMAGES
+        case kealib::kea_64int:
+            egdalType = GDT_Int64;
+            break;
+#endif
         case kealib::kea_16uint:
             egdalType = GDT_UInt16;
             break;
         case kealib::kea_32uint:
             egdalType = GDT_UInt32;
             break;
+#ifdef HAVE_64BITIMAGES
+        case kealib::kea_64uint:
+            egdalType = GDT_UInt64;
+            break;
+#endif
         case kealib::kea_32float:
             egdalType = GDT_Float32;
             break;
@@ -83,12 +95,22 @@ kealib::KEADataType GDAL_to_KEA_Type( GDALDataType egdalType )
         case GDT_Int32:
             ekeaType = kealib::kea_32int;
             break;
+#ifdef HAVE_64BITIMAGES
+        case GDT_Int64:
+            ekeaType = kealib::kea_64int;
+            break;
+#endif
         case GDT_UInt16:
             ekeaType = kealib::kea_16uint;
             break;
         case GDT_UInt32:
             ekeaType = kealib::kea_32uint;
             break;
+#ifdef HAVE_64BITIMAGES
+        case GDT_UInt64:
+            ekeaType = kealib::kea_64uint;
+            break;
+#endif
         case GDT_Float32:
             ekeaType = kealib::kea_32float;
             break;
@@ -105,18 +127,7 @@ kealib::KEADataType GDAL_to_KEA_Type( GDALDataType egdalType )
 // static function - pointer set in driver 
 GDALDataset *KEADataset::Open( GDALOpenInfo * poOpenInfo )
 {
-    bool bisKEA = false;
-    try
-    {
-        // is this a KEA file?
-        bisKEA = kealib::KEAImageIO::isKEAImage( poOpenInfo->pszFilename );
-    }
-    catch (kealib::KEAIOException &e)
-    {
-        bisKEA = false;
-    }
-
-    if( bisKEA )
+    if( Identify(poOpenInfo) )
     {
         try
         {
@@ -124,7 +135,24 @@ GDALDataset *KEADataset::Open( GDALOpenInfo * poOpenInfo )
             H5::H5File *pH5File;
             if( poOpenInfo->eAccess == GA_ReadOnly )
             {
-                pH5File = kealib::KEAImageIO::openKeaH5RDOnly( poOpenInfo->pszFilename );
+                // use the virtual driver so we can open files using
+                // /vsicurl etc
+                // do this same as libkea
+                H5::FileAccPropList keaAccessPlist =
+                    H5::FileAccPropList(H5::FileAccPropList::DEFAULT);
+                keaAccessPlist.setCache(
+                    kealib::KEA_MDC_NELMTS, kealib::KEA_RDCC_NELMTS,
+                    kealib::KEA_RDCC_NBYTES, kealib::KEA_RDCC_W0);
+                keaAccessPlist.setSieveBufSize(kealib::KEA_SIEVE_BUF);
+                hsize_t blockSize = kealib::KEA_META_BLOCKSIZE;
+                keaAccessPlist.setMetaBlockSize(blockSize);
+                // but set the driver
+                keaAccessPlist.setDriver(HDF5VFLGetFileDriver(), nullptr);
+
+                const H5std_string keaImgFilePath(poOpenInfo->pszFilename);
+                pH5File = new H5::H5File(keaImgFilePath, H5F_ACC_RDONLY,
+                                         H5::FileCreatPropList::DEFAULT,
+                                         keaAccessPlist);
             }
             else
             {
@@ -138,44 +166,49 @@ GDALDataset *KEADataset::Open( GDALOpenInfo * poOpenInfo )
 
             return pDataset;
         }
-        catch (kealib::KEAIOException &e)
+        catch (const kealib::KEAIOException &e)
         {
             // was a problem - can't be a valid file
-            return NULL;
+            CPLError( CE_Failure, CPLE_OpenFailed,
+                  "Attempt to open file `%s' failed. Error: %s\n",
+                  poOpenInfo->pszFilename, e.what() );
+            return nullptr;
+        }
+        catch (...)
+        {
+            // was a problem - can't be a valid file
+            CPLError(CE_Failure, CPLE_OpenFailed,
+                     "Attempt to open file `%s' failed. Error: unknown\n",
+                     poOpenInfo->pszFilename);
+            return nullptr;
         }
     }
     else
     {
         // not a KEA file
-        return NULL;
+        return nullptr;
     }
 }
 
 // static function- pointer set in driver
 // this function is called in preference to Open
+// can't just try and open it as could be vsi
 // 
 int KEADataset::Identify( GDALOpenInfo * poOpenInfo )
 {
-    bool bisKEA = false;
-    try
-    {
-        // is this a KEA file?
-        bisKEA = kealib::KEAImageIO::isKEAImage( poOpenInfo->pszFilename );
-    }
-    catch (kealib::KEAIOException &e)
-    {
-        bisKEA = false;
-    }
-    if( bisKEA )
-        return 1;
-    else
-        return 0;
-}
+    /* -------------------------------------------------------------------- */
+    /*      Is it an HDF5 file?                                             */
+    /* -------------------------------------------------------------------- */
+    static const char achSignature[] = "\211HDF\r\n\032\n";
 
-#ifndef HAVE_RFC40
-    // added about the same time
-    #define GDALValidateCreationOptions(a, b) TRUE
-#endif
+    if (poOpenInfo->pabyHeader == nullptr ||
+        memcmp(poOpenInfo->pabyHeader, achSignature, 8) != 0)
+    {
+        return 0;
+    }
+
+    return EQUAL(CPLGetExtension(poOpenInfo->pszFilename), "KEA");
+}
 
 // static function- pointer set in driver
 GDALDataset *KEADataset::Create( const char * pszFilename,
@@ -184,63 +217,77 @@ GDALDataset *KEADataset::Create( const char * pszFilename,
                                   char ** papszParmList  )
 {
     GDALDriverH hDriver = GDALGetDriverByName( "KEA" );
-    if( ( hDriver == NULL ) || !GDALValidateCreationOptions( hDriver, papszParmList ) )
+    if( ( hDriver == nullptr ) || !GDALValidateCreationOptions( hDriver, papszParmList ) )
     {
         CPLError( CE_Failure, CPLE_OpenFailed,
                   "Attempt to create file `%s' failed. Invalid creation option(s)\n", pszFilename);
-        return NULL;
+        return nullptr;
     }
+    
+    // This helps avoiding issues with H5File handles in a bad state, that
+    // may cause crashes at process termination
+    // Cf https://github.com/OSGeo/gdal/issues/8743
+    if (VSIFileManager::GetHandler(pszFilename) !=
+        VSIFileManager::GetHandler(""))
+    {
+        CPLError(CE_Failure, CPLE_OpenFailed,
+                 "Attempt to create file `%s' failed. /vsi file systems not "
+                 "supported\n",
+                 pszFilename);
+        return nullptr;
+    }
+        
     // process any creation options in papszParmList
     // default value
     unsigned int nimageblockSize = kealib::KEA_IMAGE_CHUNK_SIZE;
     // see if they have provided a different value
     const char *pszValue = CSLFetchNameValue( papszParmList, "IMAGEBLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nimageblockSize = atol( pszValue );
 
     unsigned int nattblockSize = kealib::KEA_ATT_CHUNK_SIZE;
     pszValue = CSLFetchNameValue( papszParmList, "ATTBLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nattblockSize = atol( pszValue );
 
     unsigned int nmdcElmts = kealib::KEA_MDC_NELMTS;
     pszValue = CSLFetchNameValue( papszParmList, "MDC_NELMTS" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nmdcElmts = atol( pszValue );
 
     hsize_t nrdccNElmts = kealib::KEA_RDCC_NELMTS;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_NELMTS" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccNElmts = atol( pszValue );
 
     hsize_t nrdccNBytes = kealib::KEA_RDCC_NBYTES;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_NBYTES" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccNBytes = atol( pszValue );
 
     double nrdccW0 = kealib::KEA_RDCC_W0;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_W0" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccW0 = atof( pszValue );
 
     hsize_t nsieveBuf = kealib::KEA_SIEVE_BUF;
     pszValue = CSLFetchNameValue( papszParmList, "SIEVE_BUF" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nsieveBuf = atol( pszValue );
 
     hsize_t nmetaBlockSize = kealib::KEA_META_BLOCKSIZE;
     pszValue = CSLFetchNameValue( papszParmList, "META_BLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nmetaBlockSize = atol( pszValue );
 
     unsigned int ndeflate = kealib::KEA_DEFLATE;
     pszValue = CSLFetchNameValue( papszParmList, "DEFLATE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         ndeflate = atol( pszValue );
 
     bool bThematic = false;
     pszValue = CSLFetchNameValue( papszParmList, "THEMATIC" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         bThematic = EQUAL(pszValue, "YES");
 
     try
@@ -249,7 +296,7 @@ GDALDataset *KEADataset::Create( const char * pszFilename,
         H5::H5File *keaImgH5File = kealib::KEAImageIO::createKEAImage( pszFilename,
                                                     GDAL_to_KEA_Type( eType ),
                                                     nXSize, nYSize, nBands,
-                                                    NULL, NULL, nimageblockSize, 
+                                                    nullptr, nullptr, nimageblockSize, 
                                                     nattblockSize, nmdcElmts, nrdccNElmts,
                                                     nrdccNBytes, nrdccW0, nsieveBuf, 
                                                     nmetaBlockSize, ndeflate );
@@ -271,12 +318,19 @@ GDALDataset *KEADataset::Create( const char * pszFilename,
 
         return pDataset;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         CPLError( CE_Failure, CPLE_OpenFailed,
                   "Attempt to create file `%s' failed. Error: %s\n",
                   pszFilename, e.what() );
-        return NULL;
+        return nullptr;
+    }
+    catch (...)
+    {
+        CPLError(CE_Failure, CPLE_OpenFailed,
+                 "Attempt to create file `%s' failed. Error: unknown\n",
+                 pszFilename);
+        return nullptr;
     }
 }
 
@@ -285,63 +339,63 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
                                 GDALProgressFunc pfnProgress, void *pProgressData )
 {
     GDALDriverH hDriver = GDALGetDriverByName( "KEA" );
-    if( ( hDriver == NULL ) || !GDALValidateCreationOptions( hDriver, papszParmList ) )
+    if( ( hDriver == nullptr ) || !GDALValidateCreationOptions( hDriver, papszParmList ) )
     {
         CPLError( CE_Failure, CPLE_OpenFailed,
                   "Attempt to create file `%s' failed. Invalid creation option(s)\n", pszFilename);
-        return NULL;
+        return nullptr;
     }
     // process any creation options in papszParmList
     // default value
     unsigned int nimageblockSize = kealib::KEA_IMAGE_CHUNK_SIZE;
     // see if they have provided a different value
     const char *pszValue = CSLFetchNameValue( papszParmList, "IMAGEBLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nimageblockSize = atol( pszValue );
 
     unsigned int nattblockSize = kealib::KEA_ATT_CHUNK_SIZE;
     pszValue = CSLFetchNameValue( papszParmList, "ATTBLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nattblockSize = atol( pszValue );
 
     unsigned int nmdcElmts = kealib::KEA_MDC_NELMTS;
     pszValue = CSLFetchNameValue( papszParmList, "MDC_NELMTS" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nmdcElmts = atol( pszValue );
 
     hsize_t nrdccNElmts = kealib::KEA_RDCC_NELMTS;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_NELMTS" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccNElmts = atol( pszValue );
 
     hsize_t nrdccNBytes = kealib::KEA_RDCC_NBYTES;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_NBYTES" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccNBytes = atol( pszValue );
 
     double nrdccW0 = kealib::KEA_RDCC_W0;
     pszValue = CSLFetchNameValue( papszParmList, "RDCC_W0" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nrdccW0 = atof( pszValue );
 
     hsize_t nsieveBuf = kealib::KEA_SIEVE_BUF;
     pszValue = CSLFetchNameValue( papszParmList, "SIEVE_BUF" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nsieveBuf = atol( pszValue );
 
     hsize_t nmetaBlockSize = kealib::KEA_META_BLOCKSIZE;
     pszValue = CSLFetchNameValue( papszParmList, "META_BLOCKSIZE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         nmetaBlockSize = atol( pszValue );
 
     unsigned int ndeflate = kealib::KEA_DEFLATE;
     pszValue = CSLFetchNameValue( papszParmList, "DEFLATE" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         ndeflate = atol( pszValue );
 
     bool bThematic = false;
     pszValue = CSLFetchNameValue( papszParmList, "THEMATIC" );
-    if( pszValue != NULL )
+    if( pszValue != nullptr )
         bThematic = EQUAL(pszValue, "YES");
 
     // get the data out of the input dataset
@@ -356,7 +410,7 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
         H5::H5File *keaImgH5File = kealib::KEAImageIO::createKEAImage( pszFilename,
                                                     GDAL_to_KEA_Type( eType ),
                                                     nXSize, nYSize, nBands,
-                                                    NULL, NULL, nimageblockSize, 
+                                                    nullptr, nullptr, nimageblockSize, 
                                                     nattblockSize, nmdcElmts, nrdccNElmts,
                                                     nrdccNBytes, nrdccW0, nsieveBuf, 
                                                     nmetaBlockSize, ndeflate );
@@ -371,7 +425,7 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
         if( !CopyFile( pSrcDs, pImageIO, pfnProgress, pProgressData) )
         {
             delete pImageIO;
-            return NULL;
+            return nullptr;
         }
 
         // close it
@@ -379,7 +433,7 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
         {
             pImageIO->close();
         }
-        catch (kealib::KEAIOException &e)
+        catch (const kealib::KEAIOException &e)
         {
         }
         delete pImageIO;
@@ -404,12 +458,12 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
 
         return pDataset;
     }
-    catch (kealib::KEAException &e)
+    catch (const kealib::KEAException &e)
     {
         CPLError( CE_Failure, CPLE_OpenFailed,
                   "Attempt to create file `%s' failed. Error: %s\n",
                   pszFilename, e.what() );
-        return NULL;
+        return nullptr;
     }
 
 }
@@ -417,11 +471,13 @@ GDALDataset *KEADataset::CreateCopy( const char * pszFilename, GDALDataset *pSrc
 // constructor
 KEADataset::KEADataset( H5::H5File *keaImgH5File, GDALAccess eAccess )
 {
+    this->m_hMutex = CPLCreateMutex();
+    CPLReleaseMutex( this->m_hMutex );
     try
     {
         // create the image IO and initilize the refcount
         m_pImageIO = new kealib::KEAImageIO();
-        m_pnRefcount = new int(1);
+        m_pRefcount = new LockedRefCount();
 
         // open the file
         m_pImageIO->openKEAImageHeader( keaImgH5File );
@@ -438,7 +494,7 @@ KEADataset::KEADataset( H5::H5File *keaImgH5File, GDALAccess eAccess )
         {
             // note GDAL uses indices starting at 1 and so does kealib
             // create band object
-            KEARasterBand *pBand = new KEARasterBand( this, nCount + 1, eAccess, m_pImageIO, m_pnRefcount );
+            KEARasterBand *pBand = new KEARasterBand( this, nCount + 1, eAccess, m_pImageIO, m_pRefcount );
             // read in overviews
             pBand->readExistingOverviews();
             // set the band into this dataset
@@ -446,14 +502,13 @@ KEADataset::KEADataset( H5::H5File *keaImgH5File, GDALAccess eAccess )
         }
 
         // read in the metadata
-        m_papszMetadataList = NULL;
+        m_papszMetadataList = nullptr;
         this->UpdateMetadataList();
 
-        // NULL until we read them in 
-        m_pGCPs = NULL;
-        m_pszGCPProjection = NULL;
+        // nullptr until we read them in 
+        m_pGCPs = nullptr;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         // ignore?
         CPLError( CE_Warning, CPLE_AppDefined,
@@ -463,33 +518,38 @@ KEADataset::KEADataset( H5::H5File *keaImgH5File, GDALAccess eAccess )
 
 KEADataset::~KEADataset()
 {
-    // destroy the metadata
-    CSLDestroy(m_papszMetadataList);
+    {
+        CPLMutexHolderD( &m_hMutex );
+        // destroy the metadata
+        CSLDestroy(m_papszMetadataList);
+        this->DestroyGCPs();
+    }
     // decrement the refcount and delete if needed
-    (*m_pnRefcount)--;
-    if( *m_pnRefcount == 0 )
+    if( m_pRefcount->DecRef() )
     {
         try
         {
             m_pImageIO->close();
         }
-        catch (kealib::KEAIOException &e)
+        catch (const kealib::KEAIOException &e)
         {
         }
         delete m_pImageIO;
-        delete m_pnRefcount;
+        delete m_pRefcount;
     }
-    this->DestroyGCPs();
-    free( m_pszGCPProjection );
+
+    CPLDestroyMutex( m_hMutex );
+    m_hMutex = nullptr;
 }
 
 // read in the metadata into our CSLStringList
 void KEADataset::UpdateMetadataList()
 {
+    CPLMutexHolderD( &m_hMutex );
     std::vector< std::pair<std::string, std::string> > odata;
     // get all the metadata
     odata = this->m_pImageIO->getImageMetaData();
-    for(std::vector< std::pair<std::string, std::string> >::iterator iterMetaData = odata.begin(); iterMetaData != odata.end(); ++iterMetaData)
+    for(auto iterMetaData = odata.begin(); iterMetaData != odata.end(); ++iterMetaData)
     {
         m_papszMetadataList = CSLSetNameValue(m_papszMetadataList, iterMetaData->first.c_str(), iterMetaData->second.c_str());
     }
@@ -511,7 +571,7 @@ CPLErr KEADataset::GetGeoTransform( double * padfTransform )
     
         return CE_None;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         CPLError( CE_Warning, CPLE_AppDefined,
                 "Unable to read geotransform: %s", e.what() );
@@ -519,18 +579,21 @@ CPLErr KEADataset::GetGeoTransform( double * padfTransform )
     }
 }
 
-// read in the projection ref
-const char *KEADataset::GetProjectionRef()
+const OGRSpatialReference* KEADataset::GetSpatialRef() const
 {
+    if( !m_oSRS.IsEmpty() )
+        return &m_oSRS;
+
     try
     {
         kealib::KEAImageSpatialInfo *pSpatialInfo = m_pImageIO->getSpatialInfo();
-        // should be safe since pSpatialInfo should be around a while...
-        return pSpatialInfo->wktString.c_str();
+        m_oSRS.SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
+        m_oSRS.importFromWkt(pSpatialInfo->wktString.c_str());
+        return m_oSRS.IsEmpty() ? nullptr: &m_oSRS;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
-        return NULL;
+        return nullptr;
     }
 }
 
@@ -552,7 +615,7 @@ CPLErr KEADataset::SetGeoTransform (double *padfTransform )
         m_pImageIO->setSpatialInfo( pSpatialInfo );
         return CE_None;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         CPLError( CE_Warning, CPLE_AppDefined,
                 "Unable to write geotransform: %s", e.what() );
@@ -561,19 +624,30 @@ CPLErr KEADataset::SetGeoTransform (double *padfTransform )
 }
 
 // set the projection
-CPLErr KEADataset::SetProjection( const char *pszWKT )
+CPLErr KEADataset::SetSpatialRef(const OGRSpatialReference* poSRS)
 {
     try
     {
         // get the spatial info and update it
         kealib::KEAImageSpatialInfo *pSpatialInfo = m_pImageIO->getSpatialInfo();
-
-        pSpatialInfo->wktString = pszWKT;
+        m_oSRS.Clear();
+        if( poSRS )
+        {
+            m_oSRS = *poSRS;
+            char* pszWKT = nullptr;
+            m_oSRS.exportToWkt(&pszWKT);
+            pSpatialInfo->wktString = pszWKT ? pszWKT: "";
+            CPLFree(pszWKT);
+        }
+        else
+        {
+            pSpatialInfo->wktString.clear();
+        }
 
         m_pImageIO->setSpatialInfo( pSpatialInfo );
         return CE_None;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         CPLError( CE_Warning, CPLE_AppDefined,
                 "Unable to write projection: %s", e.what() );
@@ -589,9 +663,15 @@ void * KEADataset::GetInternalHandle(const char *)
 
 // this is called by GDALDataset::BuildOverviews. we implement this function to support
 // building of overviews
+#ifdef HAVE_OVERVIEWOPTIONS
+CPLErr KEADataset::IBuildOverviews(const char *pszResampling, int nOverviews, const int *panOverviewList, 
+                                    int nListBands, const int *panBandList, GDALProgressFunc pfnProgress, 
+                                    void *pProgressData, CSLConstList papszOptions)
+#else
 CPLErr KEADataset::IBuildOverviews(const char *pszResampling, int nOverviews, int *panOverviewList, 
                                     int nListBands, int *panBandList, GDALProgressFunc pfnProgress, 
                                     void *pProgressData)
+#endif
 {
     // go through the list of bands that have been passed in
     int nCurrentBand, nOK = 1;
@@ -606,10 +686,10 @@ CPLErr KEADataset::IBuildOverviews(const char *pszResampling, int nOverviews, in
 
         // get GDAL to do the hard work. It will calculate the overviews and write them
         // back into the objects
-#if (GDAL_VERSION_MAJOR == 1) && (GDAL_VERSION_MINOR < 6)
-        if( GDALRegenerateOverviews( pBand, nOverviews, (GDALRasterBand**)pBand->GetOverviewList(),
-                                    pszResampling, pfnProgress, pProgressData ) != CE_None )
-#else         
+#ifdef HAVE_OVERVIEWOPTIONS
+        if( GDALRegenerateOverviewsEx( (GDALRasterBandH)pBand, nOverviews, (GDALRasterBandH*)pBand->GetOverviewList(),
+                                    pszResampling, pfnProgress, pProgressData, papszOptions) != CE_None )
+#else
         if( GDALRegenerateOverviews( (GDALRasterBandH)pBand, nOverviews, (GDALRasterBandH*)pBand->GetOverviewList(),
                                     pszResampling, pfnProgress, pProgressData ) != CE_None )
 #endif
@@ -630,8 +710,9 @@ CPLErr KEADataset::IBuildOverviews(const char *pszResampling, int nOverviews, in
 // set a single metadata item
 CPLErr KEADataset::SetMetadataItem(const char *pszName, const char *pszValue, const char *pszDomain)
 {
+    CPLMutexHolderD( &m_hMutex );
     // only deal with 'default' domain - no geolocation etc
-    if( ( pszDomain != NULL ) && ( *pszDomain != '\0' ) )
+    if( ( pszDomain != nullptr ) && ( *pszDomain != '\0' ) )
         return CE_Failure;
 
     try
@@ -641,7 +722,7 @@ CPLErr KEADataset::SetMetadataItem(const char *pszName, const char *pszValue, co
         m_papszMetadataList = CSLSetNameValue( m_papszMetadataList, pszName, pszValue );
         return CE_None;
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         return CE_Failure;
     }
@@ -650,19 +731,20 @@ CPLErr KEADataset::SetMetadataItem(const char *pszName, const char *pszValue, co
 // get a single metadata item
 const char *KEADataset::GetMetadataItem (const char *pszName, const char *pszDomain)
 {
+    CPLMutexHolderD( &m_hMutex );
     // only deal with 'default' domain - no geolocation etc
-    if( ( pszDomain != NULL ) && ( *pszDomain != '\0' ) )
-        return NULL;
+    if( ( pszDomain != nullptr ) && ( *pszDomain != '\0' ) )
+        return nullptr;
     // string returned from CSLFetchNameValue should be persistant
     return CSLFetchNameValue(m_papszMetadataList, pszName);
 }
 
-// get the whole metadata as CSLStringList
+// get the whole metadata as CSLStringList - note may be thread safety issues
 char **KEADataset::GetMetadata(const char *pszDomain)
 { 
     // only deal with 'default' domain - no geolocation etc
-    if( ( pszDomain != NULL ) && ( *pszDomain != '\0' ) )
-        return NULL;
+    if( ( pszDomain != nullptr ) && ( *pszDomain != '\0' ) )
+        return nullptr;
     // this is what we store it as anyway
     return m_papszMetadataList; 
 }
@@ -670,8 +752,9 @@ char **KEADataset::GetMetadata(const char *pszDomain)
 // set the whole metadata as a CSLStringList
 CPLErr KEADataset::SetMetadata(char **papszMetadata, const char *pszDomain)
 {
+    CPLMutexHolderD( &m_hMutex );
     // only deal with 'default' domain - no geolocation etc
-    if( ( pszDomain != NULL ) && ( *pszDomain != '\0' ) )
+    if( ( pszDomain != nullptr ) && ( *pszDomain != '\0' ) )
         return CE_Failure;
 
     int nIndex = 0;
@@ -680,7 +763,7 @@ CPLErr KEADataset::SetMetadata(char **papszMetadata, const char *pszDomain)
     try
     {
         // go through each item
-        while( papszMetadata[nIndex] != NULL )
+        while( papszMetadata[nIndex] != nullptr )
         {
             // get the value/name
             pszValue = CPLParseNameValue( papszMetadata[nIndex], &pszName );
@@ -689,7 +772,7 @@ CPLErr KEADataset::SetMetadata(char **papszMetadata, const char *pszDomain)
             nIndex++;
         }
     }
-    catch (kealib::KEAIOException &e)
+    catch (const kealib::KEAIOException &e)
     {
         return CE_Failure;
     }
@@ -706,19 +789,19 @@ CPLErr KEADataset::AddBand(GDALDataType eType, char **papszOptions)
     unsigned int nimageBlockSize = kealib::KEA_IMAGE_CHUNK_SIZE;
     unsigned int nattBlockSize = kealib::KEA_ATT_CHUNK_SIZE;
     unsigned int ndeflate = kealib::KEA_DEFLATE;
-    if (papszOptions != NULL) {
+    if (papszOptions != nullptr) {
         const char *pszValue = CSLFetchNameValue(papszOptions,"IMAGEBLOCKSIZE");
-        if ( pszValue != NULL ) {
+        if ( pszValue != nullptr ) {
             nimageBlockSize = atol(pszValue);
         }
 
         pszValue = CSLFetchNameValue(papszOptions, "ATTBLOCKSIZE");
-        if (pszValue != NULL) {
+        if (pszValue != nullptr) {
             nattBlockSize = atol(pszValue);
         }
 
         pszValue = CSLFetchNameValue(papszOptions, "DEFLATE");
-        if (pszValue != NULL) {
+        if (pszValue != nullptr) {
             ndeflate = atol(pszValue);
         }
     }
@@ -726,7 +809,7 @@ CPLErr KEADataset::AddBand(GDALDataType eType, char **papszOptions)
     try {
         m_pImageIO->addImageBand(GDAL_to_KEA_Type(eType), "", nimageBlockSize,
                 nattBlockSize, ndeflate);
-    } catch (kealib::KEAIOException &e) {
+    } catch (const kealib::KEAIOException &e) {
         return CE_Failure;
     }
 
@@ -736,12 +819,26 @@ CPLErr KEADataset::AddBand(GDALDataType eType, char **papszOptions)
     // create a new band and add it to the dataset
     // note GDAL uses indices starting at 1 and so does kealib
     KEARasterBand *pBand = new KEARasterBand(this, this->nBands, this->eAccess,
-            m_pImageIO, m_pnRefcount);
+            m_pImageIO, m_pRefcount);
     this->SetBand(this->nBands, pBand);            
 
     return CE_None;
 }
 
+OGRErr KEADataset::DeleteLayer(int iLayer)
+{
+    try
+    {
+        m_pImageIO->removeImageBand(iLayer);
+    }
+    catch (const kealib::KEAIOException &e) 
+    {
+        CPLError( CE_Warning, CPLE_AppDefined,
+                "Unable to delete a layer: %s", e.what() );
+        return OGRERR_FAILURE;
+    }
+    return OGRERR_NONE;
+}
 
 int KEADataset::GetGCPCount()
 {
@@ -749,33 +846,38 @@ int KEADataset::GetGCPCount()
     {
         return m_pImageIO->getGCPCount();
     }
-    catch (kealib::KEAIOException &e) 
+    catch (const kealib::KEAIOException &e) 
     {
         return 0;
     }
 
 }
 
-const char* KEADataset::GetGCPProjection()
+const OGRSpatialReference* KEADataset::GetGCPSpatialRef() const
 {
-    if( m_pszGCPProjection == NULL )
+    CPLMutexHolderD( &m_hMutex );
+    if( m_oGCPSRS.IsEmpty() )
     {
         try
         {
             std::string sProj = m_pImageIO->getGCPProjection();
-            m_pszGCPProjection = strdup( sProj.c_str() );
+            m_oGCPSRS.SetAxisMappingStrategy(OAMS_TRADITIONAL_GIS_ORDER);
+            m_oGCPSRS.Clear();
+            if( !sProj.empty() )
+                m_oGCPSRS.importFromWkt( sProj.c_str() );
         }
-        catch (kealib::KEAIOException &e) 
+        catch (const kealib::KEAIOException &e) 
         {
-            return NULL;
+            return nullptr;
         }
     }
-    return m_pszGCPProjection;
+    return m_oGCPSRS.IsEmpty() ? nullptr : &m_oGCPSRS;
 }
 
 const GDAL_GCP* KEADataset::GetGCPs()
 {
-    if( m_pGCPs == NULL )
+    CPLMutexHolderD( &m_hMutex );
+    if( m_pGCPs == nullptr )
     {
         // convert to GDAL data structures
         try
@@ -797,19 +899,19 @@ const GDAL_GCP* KEADataset::GetGCPs()
                 pGCP->dfGCPZ = pKEAGCP->dfGCPZ;
             }
         }
-        catch (kealib::KEAIOException &e) 
+        catch (const kealib::KEAIOException &e) 
         {
-            return NULL;
+            return nullptr;
         }
     }
     return m_pGCPs;
 }
 
-CPLErr KEADataset::SetGCPs(int nGCPCount, const GDAL_GCP *pasGCPList, const char *pszGCPProjection)
+CPLErr KEADataset::SetGCPs(int nGCPCount, const GDAL_GCP *pasGCPList, const OGRSpatialReference* poSRS)
 {
+    CPLMutexHolderD( &m_hMutex );
     this->DestroyGCPs();
-    free( m_pszGCPProjection );
-    m_pszGCPProjection = NULL;
+    m_oGCPSRS.Clear();
     CPLErr result = CE_None;
 
     std::vector<kealib::KEAImageGCP*> *pKEAGCPs = new std::vector<kealib::KEAImageGCP*>(nGCPCount);
@@ -828,16 +930,27 @@ CPLErr KEADataset::SetGCPs(int nGCPCount, const GDAL_GCP *pasGCPList, const char
     }
     try
     {
-        m_pImageIO->setGCPs(pKEAGCPs, pszGCPProjection);
+        char* pszGCPProjection = nullptr;
+        if( poSRS )
+        {
+            m_oGCPSRS = *poSRS;
+            poSRS->exportToWkt(&pszGCPProjection);
+            m_pImageIO->setGCPs(pKEAGCPs, pszGCPProjection ? pszGCPProjection : "");
+            CPLFree(pszGCPProjection);
+        }
+        else
+        {
+            m_pImageIO->setGCPs(pKEAGCPs, "");
+        }
     }
-    catch (kealib::KEAIOException &e) 
+    catch (const kealib::KEAIOException &e) 
     {
         CPLError( CE_Warning, CPLE_AppDefined,
                 "Unable to write GCPs: %s", e.what() );
         result = CE_Failure;
     }
 
-    for( std::vector<kealib::KEAImageGCP*>::iterator itr = pKEAGCPs->begin(); itr != pKEAGCPs->end(); itr++)
+    for( auto itr = pKEAGCPs->begin(); itr != pKEAGCPs->end(); itr++)
     {
         kealib::KEAImageGCP *pKEA = (*itr);
         delete pKEA;
@@ -849,7 +962,8 @@ CPLErr KEADataset::SetGCPs(int nGCPCount, const GDAL_GCP *pasGCPList, const char
 
 void KEADataset::DestroyGCPs()
 {
-    if( m_pGCPs != NULL )
+    CPLMutexHolderD( &m_hMutex );
+    if( m_pGCPs != nullptr )
     {
         // we assume this is always the same as the internal list...
         int nCount = this->GetGCPCount();
@@ -860,6 +974,6 @@ void KEADataset::DestroyGCPs()
             free( pGCP->pszInfo );
         }
         free( m_pGCPs );
-        m_pGCPs = NULL;
+        m_pGCPs = nullptr;
     }
 }
